/* TODO: move to main src file */
#include <stdlib.h>

typedef unsigned char u8_t, *u8_p;
#define	memAlloc malloc
#define memFree free
/* Based on Quake's net_loop.c::IntAlign */
#define	INT_ALIGN(value) \
		(((value) + (sizeof(int) - 1)) & (~(sizeof(int) - 1)))
#define memCopy32(dst,src,lenby4)	memcpy(dst,src,(lenby4)*4)

u8_p
vglReadPixels3(u32_t width,u32_t height)
{
	u8_t *pixels;
	u32_t w3,stride_adjust;
	u32_t half;
	void *buffer;
	u8_t *upper, *lower;
	if(!width || !height)
	{
		return NULL;
	}
	pixels=(u8_t*)memAlloc(width * height * 3);
    w3 = width * 3;
	stride_adjust = INT_ALIGN(w3)-w3;
	glPixelStorei(GL_UNPACK_LSB_FIRST,GL_FALSE);
	glPixelStorei(GL_UNPACK_ROW_LENGTH,0);
	glPixelStorei(GL_UNPACK_SKIP_ROWS,0);
	glPixelStorei(GL_UNPACK_SKIP_PIXELS,0);
	glPixelStorei(GL_UNPACK_ALIGNMENT,1);
	glReadPixels(0,0,width,height,GL_RGB,GL_UNSIGNED_BYTE,pixels);
	if(glGetError() != GL_NO_ERROR)
	{
		memFree(pixels);
		return NULL;
	}
	buffer = memAlloc(width * 3);
	half = height / 2;
	upper = pixels;
	lower = upper + (height - 1) * width * 3;
	while(half--)
	{
		memCopy32(buffer,upper,width);
		memCopy32(upper,lower,width);
		memCopy32(lower,buffer,width);
		upper += width * 3;
		lower -= width * 3;
	}
	memFree(buffer);
	return pixels;
}

GLboolean
aguSavePPM(const char *outname)
{
	unsigned int r;
	unsigned char *buffer, *row;
	FILE *out;
	GLuint width = agGetThreadWindowConfig()->width;
	GLuint height = agGetThreadWindowConfig()->height;
	buffer = malloc(3 * width * height);
	if(!buffer)
		return GL_FALSE;
	glPixelStorei(GL_UNPACK_LSB_FIRST,GL_FALSE);
	glPixelStorei(GL_UNPACK_ROW_LENGTH,0);
	glPixelStorei(GL_UNPACK_SKIP_ROWS,0);
	glPixelStorei(GL_UNPACK_SKIP_PIXELS,0);
	glPixelStorei(GL_UNPACK_ALIGNMENT,1);
	memset(buffer, 3 * width * height, 0);
	glReadPixels(0,0,width,height,GL_RGB,GL_UNSIGNED_BYTE,buffer);
	if(glGetError() != GL_NO_ERROR)
	{
		free(buffer);
		return false;
	}
	out = fopen(outname, "wb");
	if(!out)
	{
		fprintf(stderr, "aguSavePPM(): unable to open \"%s\".\n", outname);
		free(buffer);
		return GL_FALSE;
	}
	fprintf(out, "P6\n");
	fprintf(out, "%u %u\n255\n", width, height);
	row = buffer + 3 * (width - 1) * height;
	for(r = 0;r < height;++r, row -= 3 * width)
		fwrite(row, 3, width, out);
	/* fwrite(buffer, 3, width * height, out); */
	fclose(out);
	free(buffer);
	return GL_TRUE;
}

/*
 * Converts from HSV to RGB color
 */
void
aguRgbFromHsv(vec3_t rgb,vec3_t hsv)
{
	int i;
	vec_t f,p,q,t,h,s,v;
	if(IsZero(hsv[1]))
	{
		vec3Set(rgb,hsv[2],hsv[2],hsv[2]);
		return;
	}
	h = hsv[0];
	s = hsv[1];
	v = hsv[2];
	if(IsZero(h - 1))
		h = 0;
	h *= 6;
	i = (int)floor(h);
	f = h - i;
	p = v * (1 - s);
	q = v * (1 - s * f);
	t = v * (1 - s * (1 - f));
	switch(i)
	{
		case 0:
			vec3Set(rgb,v,t,p);
			break;
		case 1:
			vec3Set(rgb,q,v,p);
			break;
		case 2:
			vec3Set(rgb,p,v,t);
			break;
		case 3:
			vec3Set(rgb,p,q,v);
			break;
		case 4:
			vec3Set(rgb,t,p,v);
			break;
		case 5:
			vec3Set(rgb,v,p,q);
			break;
	}
}

/*
 * Draws arrow of length 1 along Oz axis.
 * Parameters:
 *  sides - number of subdivisions in angular direction;
 *  segs - number of subdivisions along arrow (corner is'n divided);
 *  0 < zsplit < 1 = length of arrow (1-zplit - length of corner);
 *  hscale >= 1 - relative start diameter;
 *  diameter - diameter of arrow.
 */
void
aguDrawArrow(GLuint sides, GLuint segs, GLfloat zsplit, GLfloat hscale, GLfloat diameter,
	GLfloat *cap2colorrgba)
{
	GLuint i, j;
	GLfloat x, y, z, dz = zsplit / segs;
	GLfloat phi, dphi = (GLfloat)M_2PI / sides;
	glPushMatrix();
	glScalef(diameter * 0.5, diameter * 0.5, 1);
	/* bottom cap */
	glBegin(GL_TRIANGLE_FAN);
		glVertex3f(0, 0, 0);
		for(i = 0, phi = 0;i <= sides;++i, phi -= dphi)
			glVertex3f((GLfloat)cos(phi), (GLfloat)sin(phi), 0);
	glEnd();
	/* cylinder */
	for(j = 0, z = 0;j < segs;++j, z += dz)
	{
		glBegin(GL_QUAD_STRIP);
			for(i = 0, phi = 0;i <= sides;i++, phi += dphi)
			{
				x = (GLfloat)cos(phi);
				y = (GLfloat)sin(phi);
				glVertex3f(x, y, z + dz);
				glVertex3f(x, y, z);
			}
		glEnd();
	}
	/* cap between cylinder and arrow */
	if(hscale > 1)
	{
		if(cap2colorrgba)
		{
			glPushAttrib(GL_CURRENT_BIT);
			glColor4fv(cap2colorrgba);
		}
		glBegin(GL_QUAD_STRIP);
			for(i = 0, phi = 0;i <= sides;++i, phi -= dphi)
			{
				x = (GLfloat)cos(phi);
				y = (GLfloat)sin(phi);
				glVertex3f(x, y, zsplit);
				glVertex3f(hscale * x, hscale * y, zsplit);
			}
		glEnd();
		if(cap2colorrgba)
			glPopAttrib();
	}
	/* arrow */
	glBegin(GL_TRIANGLE_FAN);
		glVertex3f(0, 0, 1);
		for(i = 0, phi = 0;i <= sides;++i, phi += dphi)
			glVertex3f(hscale * (GLfloat)cos(phi),
				hscale * (GLfloat)sin(phi), zsplit);
	glEnd();
	glPopMatrix();
}

#ifndef	SRC_ALPHA
#define	SRC_ALPHA	0.5f
#endif

#ifndef	ORIGIN_COLOR
#define	ORIGIN_COLOR	0.8f,0.8f,0,SRC_ALPHA
#endif

#ifndef	MINOR_GRID_COLOR
#define	MINOR_GRID_COLOR	0,0.4f,0,SRC_ALPHA	
#endif

#ifndef	MAJOR_GRID_COLOR
#define	MAJOR_GRID_COLOR	0.2f,0.2f,0,SRC_ALPHA	
#endif

#ifndef	BBOX_COLOR
#define	BBOX_COLOR	0.5f,0.5f,0,SRC_ALPHA
#endif

#ifndef	MINOR_GRID_LINE_WIDTH
#define	MINOR_GRID_LINE_WIDTH	0.8f
#endif

#ifndef	MAJOR_GRID_LINE_WIDTH
#define	MAJOR_GRID_LINE_WIDTH	1.7f
#endif

#ifndef MINOR_GRID_DIVS
#define MINOR_GRID_DIVS	100
#endif

#ifndef MAJOR_GRID_DIVS
#define MAJOR_GRID_DIVS	10
#endif

typedef struct agm_globals_s
{
	vec3_t	mins, maxs;
	GLfloat alpha, beta, gamma, delta;
	GLfloat deltax, deltay;
	GLfloat fmin, fcenter, fmax, ext_scale;
	vboolean auto_warp, ortho_proj;
	vboolean draw_minor_grid, draw_major_grid, draw_tripod, draw_bbox;
	vboolean smooth_lines, smooth_tripod;
}	agm_globals_t, *agm_globals_p;

agm_globals_t agm =
{
	{-1, -1, -1}, {1, 1, 1},
	0, 0, 0, 0,
	0, 0,
    2.2f, 66, 650, 3.4f,
	false, false,
	false, true, true, true,
	true,
	false
};

static void
show_agm_globals()
{
	printf("\nstatic agm_globals_t agm =\n{\n"
			"\t{%g,%g,%g}, {%g,%g,%g},\n" /* mins, maxs*/
			"\t%g, %g, %g, %g,\n" /* alpha, beta, gamma, delta */
			"\t%g, %g,\n" /* deltax, deltay */
			"\t%g, %g, %g, %g,\n" /* fmin, fcenter, fmax, ext_scale */
			"\t%s, %s,\n" /* auto_warp, ortho_proj */
			"\t%s, %s, %s, %s,\n" /* draw: minor, major, tripod, bbox */
			"\t%s, %s\n};\n", /* smooth: lines, tripod */
			agm.mins[0], agm.mins[1], agm.mins[2],
			agm.maxs[0], agm.maxs[1], agm.maxs[2],
			agm.alpha, agm.beta, agm.gamma, agm.delta,
			agm.deltax, agm.deltay,
			agm.fmin, agm.fcenter, agm.fmax, agm.ext_scale,
			agm.auto_warp ? "true" : "false", agm.ortho_proj ? "true" : "false",
			agm.draw_minor_grid ? "true" : "false", agm.draw_major_grid ? "true" : "false",
			agm.draw_tripod ? "true" : "false", agm.draw_bbox ? "true" : "false",
			agm.smooth_lines ? "true" : "false", agm.smooth_tripod ? "true" : "false");
	fflush(stdout);
}

void
aguSetSceneExtents(vec3_t mins, vec3_t maxs)
{
	vec3Scale(agm.mins, mins, 0.2f);
	vec3Scale(agm.maxs, maxs, 0.2f);
}

static void
draw_xOy_grid(vec3_t origin,vec_t width,vec_t height,unsigned num_divs)
{
	unsigned i;
	vec3_t p, dp;
	vec3Copy(p,origin);
	p[0] -= (vec_t)0.5 * width;
	p[1] -= (vec_t)0.5 * height;
	vec3Set(dp,width,height,0);
	vec3Scale(dp,dp,(vec_t)1 / num_divs);
	glBegin(GL_LINES);
		/* grid along Oy direction */
		for(i = 0;i <= num_divs;i++)
		{
			glVertex3d(p[0] + i * dp[0],p[1],p[2]);
			glVertex3d(p[0] + i * dp[0],p[1] + height,p[2]);
		}
		/* grid along Ox direction */
		for(i = 0;i <= num_divs;i++)
		{
			glVertex3d(p[0],p[1] + i * dp[1],p[2]);
			glVertex3d(p[0] + width,p[1] + i * dp[1],p[2]);
		}
	glEnd(); /* GL_LINES */
}

static void
draw_2_xOy_grids(vec3_t origin,vec3_t center,vec_t width,vec_t height)
{
	vec_t dw, dh;
	vec3_t neworg;
	if(!agm.draw_minor_grid && !agm.draw_major_grid)
		return;
	dw = width / (MAJOR_GRID_DIVS - 1);
	dh = height / (MAJOR_GRID_DIVS - 1);
	vec3Copy(neworg,center);
	neworg[0] -= (vec_t)fmod(center[0] - origin[0],dw);
	neworg[1] -= (vec_t)fmod(center[1] - origin[1],dh);
	if(agm.draw_minor_grid)
	{
		glLineWidth(MINOR_GRID_LINE_WIDTH);
		glColor4f(MINOR_GRID_COLOR);
		draw_xOy_grid(neworg,width,height,MINOR_GRID_DIVS);
	}
	if(agm.draw_major_grid)
	{
		glLineWidth(MAJOR_GRID_LINE_WIDTH);
		glColor4f(MAJOR_GRID_COLOR);
		draw_xOy_grid(neworg,width,height,MAJOR_GRID_DIVS);
	}
}

static vboolean
draw(vboolean (*draw_main)(void))
{
	vboolean result;
	vec3_t origin, center;
	vec_t scale, se;
	/* Here we can recompute bounding box */
/*
	if(scene_mesh)
	{
		int i;
		bboxClear(agm.mins, agm.maxs);
		for(i = 0;i < scene.num_vertices;i++)
			bboxAddPoint(agm.mins, agm.maxs, scene.vertex[i]);
	}
*/
	/* translate to center of all objects */
	vec3Zero(origin);
	vec3Sub(center,agm.maxs,agm.mins); /* extents */
	scale = 4.0f / vec3Abs(center);
	vec3Add(center,agm.mins,agm.maxs);
	vec3Scale(center,center,0.5f);
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	glLoadIdentity();
	glRotatef(agm.gamma,0,0,1);
	glTranslatef(0,0,-agm.fcenter);
	se = scale * agm.ext_scale; /* preserve 'scale' */
	glScalef(se,se,se); 
	glTranslatef(agm.deltax,agm.deltay,0);
	glRotatef(agm.alpha,0,1,0);
	glRotatef(agm.beta,1,0,0);
	glRotatef(agm.delta,0,0,1);
	glTranslatef(-center[0],-center[1],-center[2]);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	glEnable(GL_POINT_SMOOTH);
	glEnable(GL_POLYGON_SMOOTH);
	if(agm.smooth_lines)
	{
		glEnable(GL_LINE_SMOOTH);
		glEnable(GL_BLEND);
	}
	/* Draw coordsys-aligned grid on xOy plane */
	draw_2_xOy_grids(origin,center,10,10);
	if(agm.draw_bbox)
	{
		vec3_t mins, maxs;
		glColor4f(BBOX_COLOR);
		glLineWidth(2.0f);
		vec3Scale(mins, agm.mins, 5.0f);
		vec3Scale(maxs, agm.maxs, 5.0f);
		bboxDraw(mins, maxs);
	}
	/* Axes tripod - fixed position & scale */
	if(agm.draw_tripod)
	{
		if(agm.smooth_tripod)
		{
			glEnable(GL_POLYGON_SMOOTH);
			glEnable(GL_BLEND);
		}
		else
			glDisable(GL_BLEND);
		glMatrixMode(GL_PROJECTION);
		glPushMatrix();
		glLoadIdentity();
		glOrtho(-10,10,-10,10,-10,10);
		glMatrixMode(GL_MODELVIEW);
		glPushMatrix();
		glLoadIdentity();
		glTranslatef(8, -8, 0);
		glRotatef(agm.gamma,0,0,1);
		glRotatef(agm.alpha,0,1,0);
		glRotatef(agm.beta,1,0,0);
		glRotatef(agm.delta,0,0,1);
		glScalef(2,2,2);
		/*
		glBegin(GL_LINES);
			glColor4f(.9f,.3f,.3f,SRC_ALPHA);
			glVertex3f(0,0,0);
			glVertex3f(1,0,0);
			glColor4f(.3f,.9f,.3f,SRC_ALPHA);
			glVertex3f(0,0,0);
			glVertex3f(0,1,0);
			glColor4f(.3f,.3f,.9f,SRC_ALPHA);
			glVertex3f(0,0,0);
			glVertex3f(0,0,1);
		glEnd();
		*/
		glColor4f(.3f,.3f,.9f,SRC_ALPHA);
		aguDrawArrow(8,1,0.85,1.8,0.05,NULL);
		glRotatef(90,0,1,0);
		glColor4f(.9f,.3f,.3f,SRC_ALPHA);
		aguDrawArrow(8,1,0.85,1.8,0.05,NULL);
		glRotatef(90,-1,0,0);
		glColor4f(.3f,.9f,.3f,SRC_ALPHA);
		aguDrawArrow(8,1,0.85,1.8,0.05,NULL);
		glEnable(GL_POINT_SMOOTH);
		glEnable(GL_BLEND);
		glPointSize(10.0f);
		glBegin(GL_POINTS);
			glColor4f(ORIGIN_COLOR);
			glVertex3fv(origin);
		glEnd();
		glPopMatrix();
		glMatrixMode(GL_PROJECTION);
		glPopMatrix();
		glMatrixMode(GL_MODELVIEW);
	}
	glDisable(GL_BLEND);
	glDisable(GL_POLYGON_SMOOTH);
	glDisable(GL_LINE_SMOOTH);
	glDisable(GL_POINT_SMOOTH);
	/* Draw scene */
	glPushMatrix();
	result = draw_main();
	glPopMatrix();
	agSwap();
	return result;
}

/*
 * -----------------------------------------------------------------------
 * Camera Control
 */
static vboolean paused = true;
/*
static int autopause = 0;
*/

void
aguCorrectFrustum(void)
{
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	if(agm.ortho_proj)
		glOrtho(-1,1,-1,1,agm.fmin,agm.fmax);
	else
		glFrustum(-1,1,-1,1,agm.fmin,agm.fmax);
	glMatrixMode(GL_MODELVIEW);
}

static vboolean
handle_event_check_break(ui_event_p in,
	vboolean (*next_check_break)(ui_event_p ev))
{
	static unsigned int sscount = 0;
	char outname[256];
	switch(in->type)
	{
		case ui_key_event:
			if(in->key.press)
			{
				switch(in->key.scan)
				{
					case sc_escape:
						return 1;
					case sc_wheelup:
					case sc_minus:
					case sc_kp_subtract:
						agm.fmin *= 0.9f;
						agm.fmax *= 0.9f;
						/*
						agm.fcenter = (agm.fmin + agm.fmax) * 0.5f;
						*/
						agm.ext_scale *= 0.85f;
						aguCorrectFrustum();
						break;
					case sc_wheeldown:
					case sc_equal:
					case sc_kp_add:
						agm.fmin *= 1.1f;
						agm.fmax *= 1.1f;
						agm.ext_scale *= 1.15f;
						/*
						agm.fcenter = (agm.fmin + agm.fmax) * 0.5f;
						*/
						aguCorrectFrustum();
						break;
					case sc_space:
					case sc_kp_begin:
						agm.alpha = agm.beta = agm.deltax = agm.deltay = agm.gamma = agm.delta = 0;
						break;
					case sc_up:
						agm.deltay -= 0.1f;
						break;
					case sc_kp_up:
						agm.beta += 3;
						break;
					case sc_down:
						agm.deltay += 0.1f;
						break;
					case sc_kp_down:
						agm.beta -= 3;
						break;
					case sc_left:
						agm.deltax += 0.1f;
						break;
					case sc_kp_left:
						agm.alpha -= 3;
						break;
					case sc_right:
						agm.deltax -= 0.1f;
						break;
					case sc_kp_right:
						agm.alpha += 3;
						break;
					case sc_kp_home:
						agm.gamma += 3;
						break;
					case sc_kp_prior:
						agm.gamma -= 3;
						break;
					case sc_kp_end:
						agm.delta += 3;
						break;
					case sc_kp_next:
						agm.delta -= 3;
						break;
					case sc_w:
						agm.auto_warp = !agm.auto_warp;
						uiAutoWarp(agm.auto_warp);
						break;
	/*
					case sc_a:
						autopause = !autopause;
						break;
	*/
					case sc_s:
						sprintf(outname,"shot%03u.ppm",sscount++);
						if(aguSavePPM(outname))
							printf("Scene saved to %s\n", outname);
						break;
					case sc_p:
						paused = !paused;
						break;
					case sc_g:
						if(in->state & UI_SHIFT_MASK)
							agm.draw_major_grid = !agm.draw_major_grid;
						else
							agm.draw_minor_grid = !agm.draw_minor_grid;
						break;
					case sc_o:
						agm.ortho_proj = !agm.ortho_proj;
						agm.ext_scale *= agm.ortho_proj ? 0.05 : 20.0;
						aguCorrectFrustum();
						break;
					case sc_t:
						agm.draw_tripod = !agm.draw_tripod;
						break;
					case sc_b:
						agm.draw_bbox = !agm.draw_bbox;
						break;
					case sc_f12:
						show_agm_globals();
						break;
				}
			}
			break;
		case ui_wm_event:
			switch(in->configure.wm_event)
			{
				case ui_wm_close:
					return true;
			}
			break;
		case ui_pointer_motion_event:
			if((in->button & UI_LBUTTON_MASK) || 
				(agm.auto_warp && !(in->button & UI_RBUTTON_MASK)))
			{
				agm.alpha += in->client_dx;
				agm.beta += in->client_dy;
			}
			if(in->button & UI_RBUTTON_MASK)
			{
				agm.deltax += 0.05f * in->client_dx;
				agm.deltay -= 0.05f * in->client_dy;
			}
			break;
		default:
			break;
	}
	return next_check_break ? next_check_break(in) : false;
}

static void
agMainLoopEx(vboolean (*draw_main)(void),
			 vboolean (*next_check_break)(ui_event_p),
			 vboolean (*idle)(void*),
			 void *idle_arg)
{
	ui_event_t in;
	aguCorrectFrustum();
	for(;;)
	{
		if(idle && !paused)
		{
			if(!draw(draw_main))
				break;
			/* All extra delays should be implemented in idle() */
			if(!idle(idle_arg))
				break;
			while(uiPending())
			{
				uiNextEvent(&in);
				if(handle_event_check_break(&in, next_check_break))
					goto good_bye;
			}
		}
		else /* there is no idle(), so cool down the GPU */
		{
			if(!draw(draw_main))
				break;
			uiWait();
			while(uiPending())
			{
				uiNextEvent(&in);
				if(handle_event_check_break(&in, next_check_break))
					goto good_bye;
			}
		}
	}
good_bye:;
}

static void
agMainLoop(vboolean (*draw_main)(void),
	vboolean (*next_check_break)(ui_event_p ev))
{
	agMainLoopEx(draw_main, next_check_break, NULL, NULL);
}
